/*
 * cancan  --  mud client with telnet protocol
 *
 * written by Mattias Engdegrd (Yorick) (f91-men@nada.kth.se) 1992-95
 * Maintainers: Ilie (finnag@pvv.unit.no) (and Yorick)
 * originally inspired by the Tintin client by Peter Unold but contains no
 * Tintin code.
 * Also greatly improved upon by Vivriel and Thuzzle. As usual, all the
 * developers are in debt to countless users for suggestions and debugging.
 *
 * This program is in the public domain.
 */

#define VERSION "2.5.1d (Yorick, Vivriel, Thuzzle, Ilie 1992-95)"
#define SAVEFILEVER 1

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <memory.h>

#ifdef AIX
#include <sys/select.h>
#endif

#include "cancan.h"
#include "cmd.h"
#include "term.h"
#include "edit.h"

/* 	function declarations		*/
void read_settings();
void mainlupe();
void printmotd();
void getuserinput();
void putbackcursor();
void put_word();
void process_input();
int search_action();
int exec_action();
int printstrlen();
void mark_and_print();
void wrap_out();
void intrsig();
void termsig();
void pipesig();
void parse_line();
void parse_commands();
void denewbiefy();
void unescape();
void showprompt();
void get_remote_output();
void writetofile();

char *strdup();

extern int errno, sys_nerr;
extern char *sys_errlist[];
extern int select();
extern void bzero();
extern int strncasecmp();

/*	GLOBALS		*/

char hostname[LINELEN], portnumber[LINELEN];
char deffile[LINELEN];		/* name of definition file */
aliasnode *aliases;		/* head of alias list */
actionnode *actions;		/* head of action list */
marknode *markers;		/* head of mark list */
keynode *keydefs;		/* head of key binding list */

consess con_sess[MAXFDSCAN];	/* connections indexed by socket descriptors*/
int sockfd;			/* current socket file descriptor */
static fd_set fdset;		/* set of descriptors to select() on */

char initstr[LINELEN];		/* initial string to send on connect */
int speedwalk = 0;		/* speedwalk on? */

int linemode;			/* line mode flags (LM_* in cancan.h) */

/* for line editing */
int cols, lines;	/* screen size */
int col0;		/* input line offset (= length of prompt) */
int line0;		/* screen line where the input line starts */
char edbuf[LINELEN]; 	/* line editing buffer */
int edlen;		/* length of current input line */
int pos = 0;		/* cursor position in line */
char promptstr[LINELEN];/* entire prompt string (for redrawing) */
int verbatim = 0;	/* 1 = don't expand aliases or process semicolons */
char prefixstr[LINELEN];/* inserted in the editing buffer each time */
char inserted_next[LINELEN];	/* inserted in buffer just once */
int flashback = 0;	/* cursor is on excursion and should be put back */
int excursion;		/* where the excursion is */
int wrap = 1;		/* 1 = word wrap active */
char edattrbeg[CAPLEN];	/* starting input line attributes */
char edattrend[CAPLEN]; /* ending input line attributes */

/* ugly debug macro hack, vt100-like only */
#define NOTE(param) (printf("%c7%c[H%c[K",27,27,27),printf param,\
		     printf("%c8",27),fflush(stdout))

/* history buffer */
char *hist[MAXHIST];    /* saved history lines */
int curline = 0;        /* current history line */
int pickline = 0;	/* line to pick history from */

/* completion list */
char *words[WORDSKEPT];
int wordindex = 0;

FILE *capfile = (FILE *)NULL;	/* capture file or NULL */

int main(argc, argv)
int argc; char *argv[];
{
    char *p;
    if(argc < 2 || argc > 3) {
	fprintf(stderr,
		"usage:\t%s host port\nor\t%s definition-file\n",
		*argv, *argv);
	exit(1);
    }
  
    /* initializations */
    initstr[0] = 0;
    memset(con_sess, 0, sizeof(con_sess));
  
#ifdef MOTDFILE
    printmotd();
#endif
    init_keybindings();
    if(argc == 2) {
	/*
	 * one arg, assuming it is definition file name
	 * rules: if cancan dir is set it is searched first.
	 *        if file doesn't exist, it is created there.
	 * If a slash appears in the name, the cancan dir isn't used.
	 */
	if(strchr(argv[1], '/') == NULL && (p = getenv("CANCANDIR"))) {
	    strcpy(deffile, p);
	    if(deffile[strlen(deffile) - 1] != '/') strcat(deffile, "/");
	    strcat(deffile, argv[1]);
	    if(access(deffile, R_OK | W_OK) == -1
	       && access(argv[1], R_OK | W_OK) == 0)
		strcpy(deffile, argv[1]);
	} else
	    strcpy(deffile, argv[1]);
	if(access(deffile, R_OK | W_OK) == -1) {
	    printf("Creating %s\nHost name  :", deffile);
	    fgets(hostname, LINELEN, stdin); strtok(hostname, "\n");
	    printf("Port number:");
	    fgets(portnumber, LINELEN, stdin); strtok(portnumber, "\n");
	    default_keybindings();
	    save_settings();
	} else
	    read_settings();
    } else {
	/* two args: assume they are hostname and port number */
	strcpy(hostname, argv[1]);
	strcpy(portnumber, argv[2]);
	default_keybindings();
    }
#ifndef OMIT_PRINTVER_INIT
    printver();
#endif
    mainlupe();
    
    return 1;
}

/*
 * show current version
 */
void printver()
{
    printf("cancan ver %s (%s %s) %s\n", VERSION,
#ifdef VT100
	   "vt100-only",
#else
	   "termcaps",
#endif
#ifdef USETERMIO
	   "termio",
#else
	   "BSD sgtty",
#endif
#if __STDC__
	   "\ncompiled " __TIME__ " " __DATE__
#else
	   ""
#endif
	   );
}

#ifdef MOTDFILE
/*
 * print the message of the day if present
 */
void printmotd()
{
    char line[LINELEN];
    FILE *f = fopen(MOTDFILE, "r");
    if(f) {
	while(fgets(line, LINELEN, f))
	    fputs(line, stdout);
    }
    fflush(stdout);
}
#endif

/*
 * main loop.
 */
void mainlupe()
{
    fd_set readfds;
    int err;
    struct timeval timeoutbuf, *timeout;
    
    settermcap();
    set_terminal();
    setsigs();
    
    FD_ZERO(&fdset);
    FD_SET(0, &fdset);		/* standard input (terminal) */
    connect_sess(NULL, (*initstr ? initstr : NULL));
    
    printf("Type #help for help.\n");
    
    /* make sure the input line is always the bottom one */
    winchsig();
    gotoxy(0, lines - 1); line0 = lines - 1;
    
    /*		main lupe	*/
    
    for(;;) {
	flush_output_buffer();	/* sploooosh */

	readfds = fdset;
	if(flashback) {
	    timeoutbuf.tv_sec = 0;
	    timeoutbuf.tv_usec = FLASHDELAY;
	    timeout = &timeoutbuf;
	} else
	    timeout = (struct timeval *)NULL;
     
	do {
	    err = select(MAXFDSCAN, &readfds, NULL, NULL, timeout);
	} while(err < 0 && errno == EINTR);
	if(err < 0 && errno != EINTR) syserr("select");
	if(flashback) putbackcursor();
     
	if (FD_ISSET(0, &readfds)) {
	    err--;
	    sockfd = main_sess();
	    getuserinput();
	}
	for (sockfd = 3; err && (sockfd < MAXFDSCAN); sockfd++)
	    if (FD_ISSET(sockfd, &readfds)) {
		err--;
		get_remote_output();
	    }
    }
}

/*
 * get output from the socket and process/display it.
 */
void get_remote_output()
{
  char buffer[BUFSIZE + 10];	/* allow for a terminating \0 later */
  int got;
  got = read_socket(buffer, BUFSIZE);
  if (!got) return;
  
#ifdef DEBUGCODE
  /* debug code to see in detail what strange codes come from the server */
   {
     int i; unsigned char c;
     putchar('{');
     for(i = 0; i < got; i++) 
      {
	c = buffer[i];
	if(c < 32 || c > 127) printf("[%d]", c); else putchar(c);
      }
     printf("}\n");
   }
#endif
  
  if (con_sess[sockfd].active) 
   {  /* process only active sessions */
     if(linemode & LM_CHAR) 
      {
	/* char-by-char mode: just display output, no fuss */
	write(1, buffer, got);
      } 
     else 
      {
	/* line-at-a-time mode: process input in a number of ways */
	clearinpline();
	
	if(capfile) writetofile(capfile, buffer, got);
	process_input(buffer, got);
	
	redraw_input();
      }
   }
}

/*
 * process input from remote host:
 * detect special sequences, trigger actions, locate prompt, word-wrap,
 * print to stdout
 */
void process_input(buf, size)
char *buf; int size;
{
    char *line;
    int i, mpilen = strlen(MPI);
    int onlyprompt = 1;	/* 1 if the data packet only contained a prompt */
  
    /* process block one line at a time: */
    i = 0;
    while(i < size) {
	line = buf + i;
	/*
	 * some servers like to send NULs. Replace them with spaces, and hope
	 * for the best. (I really should cut them out entirely)
	 */
	for(; i < size && buf[i] != '\n'; i++)
	    if(buf[i] == '\0') buf[i] = ' ';
	buf[i] = '\0';	/* safe to do even if i == size, there is room */
     
	if(strncmp(MPI, line, mpilen) == 0) {
	    i += process_message(line + mpilen, size - (line + mpilen - buf))
		 + mpilen;
	} else {
	    if(i == size && strlen(line) < cols) {
		/* This is probably a prompt: strip any CR in the beginning. */
		char *p;
		for(p = line; *p == '\r'; p++);
		if(p != line) {
		    onlyprompt = 0; line = p;
		}
	   
		if(onlyprompt && !strchr(line, '\b')) /* MUME kludge */
		    strcat(promptstr, line); 
		else
		    strcpy(promptstr, line);
		col0 = printstrlen(promptstr);
	   
		printf("%s%s%s", edattrend, line, edattrbeg);
		fflush(stdout);
	   
		(void)search_action(line);
	    } else {
		onlyprompt = 0;
		if(search_action(line)) {
		    if(line0 < lines - 1) line0++;
		    if (!is_main_sess())	/* sub connection */
			printf("##%s> ", con_sess[sockfd].id);
		    if(wrap)
			wrap_out(line);
		    else {
			printf("%s", edattrend);
			mark_and_print(line);
			printf("%s", edattrbeg);
		    }
		}
		col0 = 0;
	    }
	}
	i++;
    }
}

/*
 * Determine the printed length of a string. This is sometimes less than
 * the string length since it might contain escape sequences. Treated as such
 * are "<esc> [ <non-letters> <letter>", "<esc> <non-[>", "<control-char>".
 * This is not entirely universal but covers the most common cases (i.e. ansi)
 */
int printstrlen(s)
char *s;
{
    int l;
    enum { NORMAL, ESCAPE, BRACKET } state = NORMAL;
    for(l = 0; *s; s++) {
	switch(state) {
	case NORMAL:
	    if(*s == '\033') {
		state = ESCAPE;
	    } else if(*s >= ' ') {
		l++;
	    }
	    break;

	case ESCAPE:
	    state = (*s == '[') ? BRACKET : NORMAL;
	    break;

	case BRACKET:
	    if(isalpha(*s))
		state = NORMAL;
	    break;
	}
    }
    return l;
}

/*
 * match action containing &0..&9 and return actual params in param array -
 * return 1 if matched, 0 if not
 */
int match_action (pat, src, param)
char *pat, *src, param[10][PARAMLEN];
{
  char mpat[LINELEN], *npat=0, *nsrc=0, *prm=0;
  int mbeg = 0, mword = 0, p;
  
  for (p = 0; p < 10; p++)
     param[p][0] = 0;
  
  if (*pat == '^') 
   {
     pat++;
     mbeg = 1;  /* anchor match at line start */
   }
  if (*pat == '&') mbeg = - mbeg - 1;  /* pattern starts with '&' */
  
  while (pat && *pat) 
   {
     if (*pat == '&') 
      {
	pat++;
	if (isdigit(*pat)) 
	 {
	   prm = param[*pat - '0'];
	   pat++;
	 }
	else 
	 {
	   prm = param[0];  /* support single & for back compat */
	   mword = 1;	 /* match a single word only */
	 }
      }
     
     npat = strchr(pat, '&');
     if (npat) 
      {
	strncpy(mpat, pat, npat-pat);
	mpat[npat - pat] = 0;
      }
     else
     	strcpy(mpat, pat);
     
     if (*mpat) 
      {
	nsrc = strstr(src, mpat);
	if (!nsrc) return 0;
	if (mbeg > 0) 
	 {
	   if (nsrc != src) return 0;
	   mbeg = 0;  /* reset mbeg to stop further start match */
	 }
	if (prm) 
	 {
	   strncpy(prm, src, nsrc-src);
	   prm[nsrc - src] = 0;
	 }
	}
     else if (prm) 
      {
	/* end of pattern space, strip trailing blanks */
	for (p = strlen(src) - 1; p>=0 && src[p]==' '; p--);
	strncpy(prm, src, p + 1);
	prm[p + 1] = 0;
      }
     
     /* post-processing of param */
     if (prm) 
      {
	if (mword) 
	 {
	   if (mbeg == -1) 
	    {
	      /* unanchored '&' start, take last word */
	      for (p = strlen(prm)-1; p>=0 && prm[p] == ' '; p--);
	      prm[p + 1] = 0;
	      if ((src = strrchr(prm, ' ')))
	      	 do *prm++ = *++src; while (*src);
	    } 
	   else if (!pat[1]) 
	    {
	      /* '&' at end of pattern, take first word */
	      strtok(prm, " ");
	    }
	   else
	    {
	      /* match only if param is single-worded */
	      if (strchr(prm, ' ')) return 0;
	    }
	   mbeg = mword = 0;  /* reset match flags */
	    }
	else
	   strtok(prm, ";");  /* cut off multiple commands */
      }
     
     src = nsrc + strlen(mpat);
     pat = npat;
   }
  return 1;
}

/*
 * Search for actions to trig on an input line. The line can't be trashed
 * since we want to print it on the screen later.
 * Return 1 if line should be printed, 0 if it should be suppressed.
 */
int search_action(line)
char *line;
{
    actionnode *p;
    char param[10][PARAMLEN];
  
    while(*line == '\r') line++;
    for(p = actions; p; p = p->next) {
	if(p->active && match_action(p->pattern, line, param))
	    return exec_action(param, p->commands);
    }
    return 1;
}

/*
 * print line, scanning for marked text
 * (Needs to be done: match marked text only at word boundaries)
 */
void mark_and_print(line)
char *line;
{
    marknode *np;
    char buf[LINELEN];
    char begin[CAPLEN], end[CAPLEN];

    strcpy(buf, line);
    for(np = markers; np; np = np->next) {
	char *p;
	p = strstr(buf, np->pattern);
	if(p) {
	    char *q;
	    attr_string(np->attrcode, begin, end);
	    for(q = begin; *q; q++, p++) strins(p, *q);
	    p += strlen(np->pattern);
	    for(q = end; *q; q++, p++) strins(p, *q);
	}
    }
    puts(buf);
}

/*
 * write string to stdout, wrapping words to the next line.
 */
void wrap_out(s)
char *s;
{
    char *p;
    char buf[LINELEN];
    int l;
    printf("%s", edattrend);
    while(*s == '\r') putchar(*s++);
    l = strlen(s);
    if(l == 0) {
	/* empty line */
	putchar('\n'); fflush(stdout);
	return;
    }
    while(l >= cols - col0) {
	/* need wrapping? Kill trailing spaces/carriage returns */
	for(p = s + l - 1; *p == ' ' || *p == '\r'; p--);
	if(p - s < cols - col0) {
	    p[1] = '\0'; break;
	}
	for(p = s + cols - 1 - col0; *p != ' ' && p > s; p--);
	if(p == s)
	    p = s + cols - 1 - col0 + wrapglitch;
	strncpy(buf, s, p - s); buf[p - s] = '\0';
	mark_and_print(buf);
	p = skipspace(p);
	l -= p - s;
	s = p;
    }
    if(l > 0)
	mark_and_print(s);
    printf("%s", edattrbeg);
}

/*
 * execute a trigged action - return 0 if the line should be suppressed,
 * 1 otherwise
 */
int exec_action(param, command)
char param[10][PARAMLEN], *command;
{
    char buf[LINELEN];
    int no_gag = 1;

    for( ;*command; command += 5) {
	if(strcmp("#gag", command) == 0) {
	    return 0;
	} else if(strcmp("#add", command) == 0) {
	    if(*param[0]) put_word(param[0]);
	    return no_gag;
	} else if(strncmp("#add;", command, 5) == 0) {
	    if(*param[0]) put_word(param[0]);
	} else if(strncmp("#gag;", command, 5) == 0) {
	    no_gag = 0;
	} else
	    break;
    }
    subst_param(command, buf, param);
    parse_input(buf, 0);
    return no_gag;
}

/*
 * write string to file, stripping \r
 */
void writetofile(f, str, len)
FILE *f; char *str; int len;
{
    for(; len--; str++) if(*str != '\r') putc(*str, f);
}

/*
 * read terminal input and send to parser.
 * decode keys that send escape sequences
 */
void getuserinput()
{
    unsigned char c;
    static char typed[CAPLEN];	/* chars typed so far (with partial match) */
    static nchars = 0;		/* number of them */

    /* We have 4 possible line modes:
     * line mode, local echo: line editing functions in effect
     * line more, no echo: sometimes used for passwords, no line editing
     * char mode, no echo: send a character directly, no local processing
     * char mode, local echo: extremely rare, do as above.
     */
    if(linemode & LM_CHAR) {
	int i;
	while((i = read(0, (char*)&c, 1)) < 0 && errno == EINTR);
	if(i == 1) {
	    if(write(sockfd, (char*)&c, 1) != 1)
		syserr("write single char to socket");
	    if(!(linemode & LM_NOECHO)) {
		putchar((int)c);
		fflush(stdout);
	    }
	} else
	    syserr("read from stdin");
     
    } else if(linemode & LM_NOECHO) {
	/*
	 * This is the no-echo, line-at-a-time mode usually used for
	 * password entry. This is no good, since fgets will block until
	 * the line is typed, but there is rarely anything to be read in
	 * these cases anyway. (Wouldn't it be fun to display asterisks?)
	 */
	fgets(edbuf + strlen(edbuf), LINELEN - strlen(edbuf), stdin);
	printf("\n");
	send_to_host(edbuf, strlen(edbuf));
	edbuf[0] = 0; edlen = 0;
	promptstr[0] = '\0'; col0 = 0;
	
    } else {
	/* normal mode (echo, line buffered) */
	keynode *p;

	if(read(0, (char*)&c, 1) != 1)
	    syserr("read from stdin");
	if(metakey && (c & 0x80)) { /* did meta key set the high bit? */
	    typed[nchars++] = '\033'; /* prepend esc */
	    c |= 0x7f;
	}
	/* search function key strings for match */
	typed[nchars++] = c; typed[nchars] = 0;
	for(p = keydefs;
	    p && strncmp(typed, p->sequence, nchars);
	    p = p->next);
	if(p) {
	    if(strlen(p->sequence) == nchars) {
		p->funct(p->call_data);
		typed[nchars = 0] = 0;
	    }
	} else {
	    insert_char((int)c);
	    typed[nchars = 0] = 0;
	}
    }
}


/*
 * Flash cursor at parentheses that matches c inserted before current pos
 */
void flashparen(c)
int c;
{
    int lev, i;
    for(i = pos - 1, lev = 0; i >= 0; i--)
	if(ISRPAREN(edbuf[i])) {
	    lev++;
	} else if(ISLPAREN(edbuf[i])) {
	    lev--;
	    if(lev == 0)
		if(LPAREN(c) == edbuf[i])
		    break;
		else
		    i = -1;
	}
    if(i >= 0) {
	optmove(CURLINE(pos), CURCOL(pos), CURLINE(i), CURCOL(i));
	fflush(stdout);
	flashback = 1;
	excursion = i;
    }
}

/*
 * put cursor back where it belongs
 */
void putbackcursor()
{
    optmove(CURLINE(excursion), CURCOL(excursion), CURLINE(pos), CURCOL(pos));
    fflush(stdout);
    flashback = 0;
}

/*
 * put string in history at current position
 * (string is assumed to be trashable)
 */
void put_in_history(str)
char *str;
{
    char *p;
    if(hist[curline]) free(hist[curline]);
    hist[curline] = strdup(str);
    if(++curline == MAXHIST) curline = 0;
  
    /* split into words and put into completion list */
    for(p = strtok(str, DELIM); p; p = strtok((char*)NULL, DELIM)) {
	if(strlen(p) >= 3 && p[0] != '#') /* no commands/short words */
	    put_word(p);
    }
}

/*
 * put word in word completion list
 */
void put_word(word)
char *word;
{
  if(words[wordindex]) free(words[wordindex]);
  words[wordindex] = strdup(word);
  if(++wordindex == WORDSKEPT) wordindex = 0;
}

/*
 * match and complete a word referring to the word list
 */
void complete_word()
{
    int i, j;
  
    if(pos) {
	/* find word start */
	for(j = pos; j > 0 && strchr(DELIM, edbuf[j - 1]) == NULL; j--);
	if(j == pos) return;
     
	if(edbuf[j] == '#' && pos - j > 1) {
	    /* try to complete built-in command */
	    char *p;
	    for(i = 0; (p = commands[i].name); i++) {
		if(strncmp(p, edbuf + j + 1, pos - j - 1) == 0)
		    break;
	    }
	    if(p)
		insertchars(p + pos - j - 1, strlen(p + pos - j) + 1);
	} else {
	    /* scan word list for match */
	    for(i = wordindex - 1; i != wordindex; i--) {
		if(i < 0) i += WORDSKEPT;
		if(words[i] == NULL) return;
		if(strncasecmp(words[i], edbuf + j, pos - j) == 0) break;
	    }
	    if(i != wordindex)
		insertchars(words[i] + pos - j, strlen(words[i] + pos - j));
	}
    }
}

/*
 * match and complete entire lines backwards in history
 */
void complete_line()
{
  int i;
  
  if(edlen) {
    for(i = curline - 1; i != curline; i--) 
     {
       if(i < 0) i = MAXHIST - 1;
       if(hist[i] && strncmp(edbuf, hist[i], edlen) == 0) break;
     }
    if(i != curline) 
     {
       clearinpline();
       strcpy(edbuf, hist[i]);
       pos = edlen = strlen(edbuf);
       redraw_input();
     }
  }
}

#ifdef NEWBIE
/*
 * Some people just can't write prose. It's horrible, but here is the
 * fix we have been waiting for.
 */
void denewbiefy()
{
  char *p, *q, *r;
  char buf[LINELEN];
  int n;
  
  strcpy(buf, edbuf);
  p = buf; r = edbuf;
  while(*p) 
   {
     while(*p && !isalnum(*p)) *r++ = *p++;
     q = p;
     while(*p && isalnum(*p)) p++;
     if(*p || *q) 
      {
	n = p - q;
	if(strncasecmp(q, "u", n) == 0) 
	 {
	   strcpy(r, "you"); r += 3;
	 }
	else if(strncasecmp(q, "ur", n) == 0) 
	 {
	   strcpy(r, "your"); r += 4;
	 }
	else if(strncasecmp(q, "urs", n) == 0) 
	 {
	   strcpy(r, "yours"); r += 5;
	 }
	else if(strncasecmp(q, "r", n) == 0) 
	 {
	   strcpy(r, "are"); r += 3;
	 }
	else if(strncasecmp(q, "4", n) == 0) 
	 {
	   strcpy(r, "for"); r += 3;
	 }
	else if(strncasecmp(q, "2", n) == 0) 
	 {
	   strcpy(r, "to"); r += 2;
	 }
	else if(strncasecmp(q, "b", n) == 0) 
	 {
	   strcpy(r, "be"); r += 2;
	 } 
	else if(strncasecmp(q, "b4", n) == 0) 
	 {
	   strcpy(r, "before"); r += 6;
	 }
	else 
	 {
	   strcpy(r, q); r += n;
	 }
      }
   }
  *r = '\0';
  edlen = r - edbuf;
}
#endif

/*
 * put string into input line (will be redrawn next time)
 */
void to_input_line(str)
char *str;
{
    strcpy(inserted_next, str);
}

/* editor utility functions: */

/* insert one character first in a string */
void strins(s, c)
char *s; char c;
{
  char d;
  for(; *s; d = *s, *s++ = c, c = d);
  *s++ = c; *s = 0;
}

/* delete first character of a string */
void strdel(s)
char *s;
{
  while(*s) *s = s[1], s++;
}

/*
 * send a string to the main session on the remote host
 */
void send_to_main_host(str, len)
char *str;
int len;
{
    int osockfd = sockfd;	/* backup current socket fd */
    sockfd = main_sess();
    send_to_host(str, len);
    sockfd = osockfd;
}

/*
 * terminate cancan as cleanly as possible
 */
void exitcancan()
{
  printf("#cancan stopped dancing.%s\n", edattrend);
  reset_terminal();
  exit(0);
}
 
/*
 * set up our signal handlers
 */
void setsigs()
{
  signal(SIGCHLD, childdeathsig);
  signal(SIGINT, intrsig);
  signal(SIGTERM, termsig);
  signal(SIGPIPE, pipesig);
  signal(SIGWINCH, winchsig);
}

void pipesig()
{
  printf("\n#no tobacco error.\n");
  exitcancan();
}

void intrsig()
{
  printf("\n#interrupted.\n");
  exitcancan();
}

void termsig()
{
  printf("\n#termination signal.\n");
  exitcancan();
}

/*
 * suspend ourselves
 */
void suspend_cancan()
{
    printf("%s", edattrend);
    if(signal(SIGTSTP, SIG_IGN) == SIG_DFL) {
	reset_terminal();
	signal(SIGTSTP, SIG_DFL);
	printf("\nSuspended\n");
	if(kill(0, SIGTSTP) == -1) syserr("susp cancan");
	set_terminal();
	winchsig();		/* in case size changed meanwhile */
	gotoxy(0, lines - 1);	/* so we know where the cursor is */
	line0 = lines - 1;
	printf("%s%s%s", edattrend, promptstr, edattrbeg);
	redraw_input();
    } else
	printf("\n#I don't think your shell has job control.\n");
}

/*
 * print system call error message and terminate
 */
void syserr(msg)
char *msg;
{
  fprintf(stderr, "cancan system call error: %s (%d", msg, errno);
  if(errno > 0 && errno < sys_nerr)
     fprintf(stderr, ": %s", sys_errlist[errno]);
  fprintf(stderr, ")\n");
  reset_terminal();
  exit(1);
}

/*
 * parse input from user: chop the input string into semicolon-separated
 * strings and calls parse_line for each. This function is not allowed to
 * trash cmdline.
 * silent = 1 if the line should not be echoed, 0 otherwise.
 */
void parse_input(cmdline, silent)
char *cmdline; int silent;
{
  int esc;
  
  char cmd[BUFSIZE];
  char *src, *dst;
  
  /*
   * Don't unescape escaped characters now - leave them to be unescaped
   * just before they are sent instead, to avoid doing it several times.
   */
  for(esc = 0, src = cmdline, dst = cmd; *src; src++) 
   {
     if(!esc) 
      {
	if(*src == CMDSEP) 
	 {
	   *dst++ = '\0';
	   parse_line(cmd, silent);
	   dst = cmd;
	 }
	else 
	 {
	   if(*src == '\\') esc = 1;
	   *dst++ = *src;
	 }
      }
     else 
      {
	*dst++ = *src;
	esc = 0;
      }
   }
  *dst = '\0';
  parse_line(cmd, silent);
}

/*
 * execute walk if word is valid [speed]walk sequence -
 * return 1 if walked, 0 if not
 */
int parse_walk (word)
char *word;
{
    long strtol();
    char buf[16];
    int n = strlen(word);
  
    if(!n || (n > 1 && !speedwalk)
	|| !strchr("neswud", word[n - 1])
	|| strspn(word, "neswud0123456789") != n)
	return 0;

    strcpy(buf, " \n");
    while (*word) {
	if(isdigit(*word))
	    n = strtol(word, &word, 10);
	else
	    n = 1;
	while(n--) {
	    buf[0] = *word;
	    send_to_host(buf, 2);
	    if(is_main_sess())
		add_dir_to_map(*word);
	}
	word++;
    }
    return 1;
}

/*
 * split str into words separated by DELIM, and place in param[1]..param[9] -
 * the whole str is put in param[0]
 */
void split_words (str, param)
char *str, param[10][PARAMLEN];
{
  int i;
  char *prm;
  
  strcpy(param[0], str);
  for (i=1; i < 10; i++) 
   {
     prm = param[i];
     while (*str && strchr(DELIM, *str))
     	str++;  /* skip span of DELIM */
     while (*str && !strchr(DELIM, *str))
     	*prm++ = *str++;
     *prm = 0;
   }
}

/*
 * parse one command line: look for # commands and expand aliases.
 * silent = 1 if the line should not be echoed, 0 otherwise.
 */
void parse_line(line, silent)
char *line; int silent;
{
    aliasnode *p;
    char cmd[LINELEN], buf[BUFSIZE], param[10][PARAMLEN];
    char *s, *d;
    int l;
  
    /* pick out the first word = command or alias */
    for(s = line, d = cmd; *s != ' ' && *s;) *d++ = *s++;
    *d = '\0';
    s = skipspace(s);
  
    if (cmd[0] == '#' && cmd[1] == '#') {
	/* send to other session */
	if (!(sockfd = find_sess(cmd + 2)))
	    return;
	/* reform cmd and args */
	for(d = cmd; *s != ' ' && *s;) *d++ = *s++;
	*d = '\0';
	s = skipspace(s);
    }
  
    p = *lookup_alias(cmd);
    if(p) {
	split_words(s, param);  /* split s into words */
	if(!subst_param(p->subst, buf, param) && *param[0]) {
	    strcat(buf, " ");
	    strcat(buf, param[0]);
	}
	parse_input(buf, 0);
    } else if(*cmd == '#') {
	unescape(s);
	parse_commands(cmd + 1, s);
    } else if(*s || !parse_walk(cmd)) {
	unescape(cmd);
	unescape(s);
	if(*s) {
	    strcat(cmd, " ");
	    strcat(cmd, s);
	}
	if(!silent) printf("[%s]\n", cmd);
	promptstr[0] = '\0'; col0 = 0;
	l = strlen(cmd);
	cmd[l] = '\n';
	send_to_host(cmd, l + 1);
    }
}

/*
 * remove escapes (backslashes) from a string
 */
void unescape(s)
char *s;
{
  char *p;
  int esc = 0;
  for(p = s; *s; s++)
     if(esc) 
      {
	*p++ = *s;
	esc = 0;
      }
     else 
      {
	if(*s == '\\')
	   esc = 1;
	else
	   *p++ = *s;
      }
  *p = '\0';
}

/*
 * parse cancan's own commands
 */
void parse_commands(command, arg)
char *command, *arg;
{
  int i;
  
  /* disallow special commands on subsidiary connections */
  if (!is_main_sess()) 
   {
     printf("#command not allowed on subsidiary connection\n");
     showprompt();
     return;
   }
  
  /* assume output will be enough to make input line = last screen line */
  line0 = lines - 1;
  if(isdigit(*command)) 
   {
     i = atoi(command);
     if(i > 0)
     	while(i--) parse_input(arg, 1);
     else 
      {
	printf("#bogus repeat count\n");
	showprompt();
      }
   } 
  else
   {
     for(i = 0; commands[i].name; i++)
     	if(strcmp(command, commands[i].name) == 0) 
         {
	  if(commands[i].funct) 
	    {
	      printf("%s", edattrend);
	      if((*commands[i].funct)(arg))
	      	 showprompt();
	    }
	   return;
	 }
     printf("#unknown cancan command '%s'\n", command);
     showprompt();
   }
}

/*
 * redisplay the prompt
 */
void showprompt()
{
  printf("%s%s%s", edattrend, promptstr, edattrbeg);
  fflush(stdout);
}

/*
 * cancel an editing session; does not free anything
 * (the child death signal handler will remove the session from the list)
 */
void cancel_edit(sp)
editsess *sp;
{
  char buf[LINELEN];
  char keystr[16];
  
  if(kill(sp->pid, SIGKILL) < 0)	/* Editicide */
  syserr("kill editor child");
  printf("#killed %s (%d)\n", sp->descr, sp->key);
  sprintf(keystr, "C%d\n", sp->key);
  sprintf(buf, "%sE%d\n%s", MPI, (int) strlen(keystr), keystr);
  send_to_main_host(buf, strlen(buf));
  sp->key = -1;	/* mark the child as cancelled */
}

/*
 * substitute &0..&9 in a string (unless & is escaped with backslash)
 */
int subst_param(src, dst, subst)
char *src, *dst, subst[10][PARAMLEN];
{
  int subs = 0;
  int escaped = 0, i;
  
  while(*src) 
   {
     if(!escaped && *src == '&') 
      {
	src++;
	if (*src == '&') 
	 {  /* delayed substitution */
	   do *dst++ = *src++; while (*src == '&');
	   if (isdigit(*src)) *dst++ = *src++;
	 }
	else 
	 {
	   if (isdigit(*src)) 
	    {
	      i = *src - '0';
	      src++;
	    }
	   else
	      i = 0;  /* support single & for back compat */
	   strcpy(dst, subst[i]);
	   dst += strlen(subst[i]);
	   subs++;
	 }
      } 
     else
      {
	if(*src == '\\')
	   escaped ^= 1;
	else
	   escaped = 0;
	*dst++ = *src++;
      }
   }
  *dst = '\0';
  return subs;
}

/*
 * return pointer to next non-blank char
 */
char *skipspace(p)
char *p;
{
    while(*p == ' ') p++;
    return p;
}

/*
 * read definitions from file
 */
void read_settings()
{
    char buf[LINELEN];
    FILE *f = fopen(deffile, "r");
    char *p;
    int version = 0;		/* save file version */
    
    if(!f) syserr("fopen");
    while(fgets(buf, LINELEN, f) != NULL) {
	p = buf;
	while(*p != ' ' && *p) p++;
	if(*p) {
	    *p++ = '\0';
	    strtok(p, "\n");	/* strip lf */
	    if(strcmp(buf, "#savefile") == 0) {
		version = atoi(p);
		if(version > SAVEFILEVER) {
		    printf("Warning: this cancan version is too old!\n");
		}
	    } else if(strcmp(buf, "#alias") == 0) {
		parse_alias(p, 1);
	    } else if(strcmp(buf, "#host") == 0) {
		sscanf(p, "%s %s", hostname, portnumber);
	    } else if(strcmp(buf, "#init") == 0) {
		if(!strtok(p, "=\n"))
		    printf("bad definition file format\n");
		else
		    strcpy(initstr, p);
	    } else if(strcmp(buf, "#action") == 0) {
		if(version >= 1)
		    parse_action(p, 1);
		else {
		    /* old syntax, for compatibility */
		    if(!strtok(p, "=\n"))
			printf("bad definition file format\n");
		    else
			add_anonymous_action(p, strtok(NULL, "\n"), 1);
		}
	    } else if(strcmp(buf, "#mark") == 0) {
		parse_marker(p, 1);
	    } else if(strcmp(buf, "#bind") == 0) {
		parse_bind_from_file(p);
	    } else {
		printf("Invalid definition file command '%s'\n", buf);
	    }
	} else
	    printf("bad definition file format\n");
    }
    fclose(f);
    if(version < 1)
	default_keybindings();
}

/*
 * save settings in definition file
 */
void save_settings()
{
    FILE *f;
    aliasnode *alp;
    actionnode *acp;
    marknode *mp;
    keynode *kp;
  
    if(!*deffile) return;
    if(!(f = fopen(deffile, "w"))) {
	printf("#error: can't write to definition file %s\n", deffile);
	return;
    }
    fprintf(f, "#savefile %d\n", SAVEFILEVER);
    fprintf(f, "#host %s %s\n", hostname, portnumber);
    if(*initstr)
	fprintf(f, "#init %s\n", initstr);
    for(alp = aliases; alp; alp = alp->next)
	fprintf(f, "#alias %s=%s\n", alp->name, alp->subst);
    for(acp = actions; acp; acp = acp->next)
	fprintf(f, "#action %c%s \"%s\"=%s\n", (acp->active) ? '+' : '-',
		acp->label, acp->pattern, acp->commands);
    for(mp = markers; mp; mp = mp->next)
	fprintf(f, "#mark %s=%s\n", mp->pattern, attr_name(mp->attrcode));
    for(kp = keydefs; kp; kp = kp->next)
	if(kp->funct == key_run_command)
	    fprintf(f, "#bind %s %s=%s\n", kp->name, seq_name(kp->sequence),
		    kp->call_data);
    fclose(f);
}

/*
 * if alias recording is active, record given line in that alias
 */
void record_line(line)
char *line;			/* mustn't be trashed! */
{
    if(recording_name && line[0] != '#') { /* don't record #commands */
	aliasnode **np = lookup_alias(recording_name);
	char new_line[LINELEN];
	if(strcmp(line, recording_name) == 0) {
	    printf("#recording to the same alias! (recording turned off)");
	    free(recording_name);
	    recording_name = NULL;
	    return;
	}
	if(*np == NULL) {
	    add_aliasnode(recording_name, line);
	} else {
	    if(strlen((*np)->subst) + strlen(line) + 1 >= LINELEN - 16) {
		/* too long, must be split */
		char new_name[LINELEN];
		int i = 0;		/* try name1, name2, ... */
		do {
		    sprintf(new_name, "%s%d", recording_name, ++i);
		} while(*lookup_alias(new_name));
		/* now we have a fresh new name */
	    add_aliasnode(new_name, (*np)->subst);
		strcpy(new_line, new_name);
	    } else
		strcpy(new_line, (*np)->subst);
	    strcat(new_line, ";");
	    strcat(new_line, line);
	    free((*np)->subst);
	    (*np)->subst = strdup(new_line);
	}
	save_settings();
    }
}


#ifdef NOSTRDUP
/*
 * some old libraries don't have strdup(), so...share and enjoy. 
 */
char *strdup(s)
char *s;
{
  char *p = (char*)malloc(strlen(s) + 1);
  if(p) strcpy(p, s);
  return p;
}
#endif


/*
 * Below are multiple-session support functions:
 *
 * note on con_sess[MAXFDSCAN]:
 *	con_sess[0].active holds total no. of connections
 *	con_sess[1].active holds main socket descriptor
 */

/*
 * return TRUE if current session being processed is main connection
 */
int is_main_sess ()
{
  return (sockfd == con_sess[1].active);
}

/*
 * return main session sockfd
 */
int main_sess ()
{
  return con_sess[1].active;
}

/*
 * return session's index or socket descriptor given id,
 * or 0 if null or invalid id is given
 */
int find_sess (id)
char *id;
{
  int i, n;
  
  for (i=3, n=con_sess[0].active; n; i++)
     if (con_sess[i].id) 
      {
	n--;
	if (!strcmp(con_sess[i].id, id))
	   return i;
      }
  printf("#no such connection\n");
  return 0;
}

/*
 * connect another session
 */
void connect_sess (id, initstr)
char *id;	/* session id - NULL if main session */
char *initstr;	/* optional init string */
{
  /* dial the number by moving the right index in small circles */
  if ((sockfd = doconnect(hostname, atoi(portnumber))) < 0) 
   {
     if (!id)  /* main session */
        syserr("#connect: host not responding");
     else
     	printf("host not responding or too many open connections\n");
   }
  else 
   {
     FD_SET(sockfd, &fdset);	 /* add socket to select() set */
     con_sess[0].active++;
     if (!id) 
      {  /* main session */
	con_sess[1].active = sockfd;  /* save main socket fd */
	con_sess[sockfd].active = 1;
	con_sess[sockfd].id = "main";
      } 
     else
      {
	if (!initstr || !*initstr)
	   con_sess[sockfd].active = 1;  /* active only if no init */
	con_sess[sockfd].id = (char *)malloc(strlen(id) + 1);
	if (!con_sess[sockfd].id) syserr("malloc");
	strcpy(con_sess[sockfd].id, id);
      }
     if (initstr)
     	parse_input(initstr, 0);
   }
}

/*
 * disconnect session
 */
void disconnect_sess (id)
char *id;
{
  int sfd;
  
  if (id)
   {  /* #disconnect cmd */
     if (!(sfd = find_sess(id)))
     	return;
     close(sfd);
   } 
  else
     sfd = sockfd;  /* connection closed by remote host */
  
  printf("#Connection on %s closed.\n", con_sess[sfd].id);
  if (sfd == con_sess[1].active)  /* main connection closed */
  exitcancan();
  
  FD_CLR(sfd, &fdset);
  con_sess[0].active--;
  con_sess[sfd].active = 0;
  free(con_sess[sfd].id);
  con_sess[sfd].id = 0;
}

/*
 * toggle output display from another session
 */
void snoop_sess (id)
char *id;
{
    int sfd;
  
    sfd = find_sess(id);
    if(sfd)
	con_sess[sfd].active ^= 1;
}
